# -*- coding: Latin-1 -*-
## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
# DEVS.py --- Classes and Tools for 'Classic' DEVS Model Specification
#                     --------------------------------
#                            Copyright (c) 2000
#                          Jean-Sbastien  BOLDUC
#                             Hans  Vangheluwe
#                            Yentl Van Tendeloo
#                       McGill University (Montral)
#                     --------------------------------
# Version 1.0                                        last modified: 01/11/01
# Version 1.0.1                                      last modified: 04/04/05
# MODIFIED: 
#  - The default "AtomicDEVS.timeAdvance" method returns 
#    the object "INFINITY" which stands for infinity
#  - Deal with DeprecationWarning
# Version 1.0.3                                      last modified: 16/11/05
# MODIFIED: cosmetic changes 
#  - Use True/False instead of 1/0
#  - Cleanup of terminology
# Version 1.1                                        last modified: 21/11/12
# MODIFIED:
#  - Various optimisations for increased simulation speed
#
## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
#
# GENERAL NOTES AND REMARKS:
#
# Atomic- and coupled-DEVS specifications adapted from:
#       B.P.Zeigler, H. Praehofer, and Tag Gon Kim,
#       ``Theory of modeling and simulation: Integrating
#       Discrete Event and Continuous Complex Dynamic Systems '',
#       Academic Press, 2000
#
## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##

##  GLOBAL VARIABLES AND FUNCTIONS
## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##

from devs_exceptions import *

# {\sl Verbose \/} output (to stdout) when the {\tt __VERBOSE__} 
# switch evaluates to True.
# Similar to {\tt #define _DEBUG_} convention in {\tt C++}.
__VERBOSE__ = True 


##  CLASS HIERARCHY 
## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##

class BaseDEVS:
  """Abstract base class for {\tt AtomicDEVS} and {\tt CoupledDEVS} classes.
  
  This class provides basic DEVS attributes and query/set methods.
  """
  
  ###
  def __init__(self, name):
    """Constructor.
    """
    
    # Prevent any attempt to instantiate this abstract class
    if self.__class__ == BaseDEVS:
      raise DEVSException ("Cannot instantiate abstract class '%s' ... " % (self.__class__.__name__), 0)
    # {\sl General DEVS Attributes\/} --- related to atomic- and coupled-DEVS'
    # implementation:
    #
    # * {\tt __doc__} is the (optional) {\sl standard documentation string},
    #   whose content is left to the modeler (undeclared here, as
    #   it is a default class attribute in Python);
    # * {\tt parent} is a {\sl reference to the parent coupled-DEVS\/} in the
    #   hierarchical representation of the model. Declared here only for
    #   reference, as the default value gets overwritten when a DEVS is being
    #   added to a model using {\tt CoupledDEVS}'s {\tt addSubModel} method
    #   ({\tt parent} remains {\tt None} for a root-DEVS);
    # * {\tt myID} is a {\sl unique identification string\/}, which consists
    #   of the 'type' letter '{\tt A}' or '{\tt C}' (Atomic or Coupled DEVS),
    #   followed by the 'index' of the object (the number of objects {\sl of
    #   the same type\/} that will have been instantiated after the present
    #   one, {\it i.e.}, $\hbox{index}\in [1,\,+\infty[$\/). Once again
    #   declared only for reference, as the default value gets overwritten
    #   when a descriptive class is instantiated;
    # * {\tt condFlag} is the {\sl condition flag\/}, indicating whether the
    #   DEVS is {\sl active\/} (by default) or {\sl passive\/} --- True or False,
    #   respectively (for future use, in variable structure DEVS).

    self.fullName = None
    self.parent = None
    self.myID   = None
    self.condFlag = True 

    # Allow the specification of a model {\tt name} 
    # Note how the name of the instance cannot be used
    # as there might be multiple references to the same model.
    self.name = name
    
    # {\sl Specific Attributes\/} --- related to atomic- and coupled-DEVS'
    # specification:
    # * {\tt IPorts} and {\tt OPorts} are lists of references to the DEVS'
    #   input and output ports;
    # * {\tt myInput} and {\tt myOutput} are the input and output dictionaries,
    #   each of the form $\{${\sl Port_reference : message_on_port\/}$\}$.
    #   Messages are sent and retrieved using the {\tt poke} and {\tt peek} methods
    #   respectively, and the actual message 'delivery' from a DEVS to another 
    #   is left to the simulator.
    self.IPorts  = []  
    self.OPorts   = []
    self.myInput = {}  
    self.myOutput = {}

  ###
  def type(self):
    """Returns the 'type' of the object: {\tt 'ATOMIC'} or {\tt 'COUPLED'}.
    """
    if self.myID[0] == 'A':
      return 'ATOMIC'
    elif self.myID[0] == 'C':
      return 'COUPLED'
    # Else, invalid type:
    return None
      
  ###
  def addInPort(self, name=None):
    """Add an {\sl input port\/} to the DEVS model.
    
    {\tt addInPort} and {\tt addOutPort} are the {\sl only\/} proper way to
    add I/O ports to DEVS. As for the {\tt CoupledDEVS.addSubModel} method, calls
    to {\tt addInPort} and {\tt addOutPort} can appear in any DEVS'
    descriptive class constructor, or the methods can be used with an
    instantiated object.
    
    The methods add a reference to the new port in the DEVS' {\tt IPorts} or
    {\tt OPorts} attributes and set the port's {\tt hostDEVS} attribute. The
    modeler should typically add the returned reference the local dictionary
    (to be able to refer to the port for peek/poke-ing later).    
    """

    # Instantiate an input port:
    port = Port(isInput=True, name=name) 
    
    self.IPorts.append(port)
    port.hostDEVS = self
    return port
      
  ###
  def addOutPort(self, name=None):
    """Add an {\sl output port\/} to the DEVS model.
    
    See comments for {\tt addInPort} above.
    """

    # Instantiate an output port:
    port = Port(isInput=False, name=name)
        
    self.OPorts.append(port)
    port.hostDEVS = self        
    return port  
    
  ###
  def getModelName(self):
    """Get the local model name
    """
    if self.name == None:
      return self.myID
    else:
      return self.name

  ###
  def getModelFullName(self):
    """Get the full model name, including the path from the root
    """
    if self.fullName == None:
        if self.parent == None:
                self.fullName = self.getModelName()
        else:
                self.fullName = self.parent.getModelFullName() + "." + self.getModelName()
    return self.fullName

  ###

  def setHierarchy(self, hierarchy):
        hierarchy.append(self.auto_id)
        #print("Set hierarchy to " + str(hierarchy))
        self.hierarchy = hierarchy

  ###
  def getPortName(model, port):
    if not isinstance(port, Port):
      return None
    for attr in dir(model):
      if getattr(model, attr) == port:
        return attr

#    ===================================================================    #

class AtomicDEVS(BaseDEVS):
  """Abstract base class for all atomic-DEVS descriptive classes.
    
  An atomic-DEVS is described as a structure... (to be completed)
  """
  
  # {\tt AtomicIDCounter} is a static counter used for the 'index' part of
  # atomic-DEVS' {\tt myID} attribute. Incremented with every instantiation
  # of an atomic-DEVS' descriptive class, unaffected by the deletion of those
  # classes.
  AtomicIDCounter = 0
  
  ###  
  def __init__(self, name=None):
    """Constructor.
    """

    # Prevent any attempt to instantiate this abstract class
    if self.__class__ == AtomicDEVS:
      raise DEVSException("Cannot instantiate abstract class '%s' ... " % (self.__class__.__name__), 0)

    # The minimal constructor shall {\sl first\/} call the superclass
    # ({\it i.e.}, {\tt BaseDEVS}') constructor.
    BaseDEVS.__init__(self, name)
    
    # Increment {\tt AtomicIDCounter\/} and setup instance's {\tt myID\/}
    # attribute.
    AtomicDEVS.AtomicIDCounter += 1
    self.myID = "A%d" % AtomicDEVS.AtomicIDCounter
    
    # {\sl Specific Attributes\/} --- related to atomic-DEVS' specification:
    # (to be completed)
    self.elapsed = 0.0 
    self.state = None

    self.peek_ok = {}
    self.poke_ok = {}

  ###
  def poke(self, port, ev):
    """Outputs event {\tt ev} to output port {\tt port}.
    
    This merely amounts to adding a new entry into dictionary
    {\tt myOutput}. Note that the dictionary must be cleared before poking
    any port. This is the simulator's responsability.
    """

    # See if it is already checked
    if port not in self.poke_ok:
     if port.type() != 'OUTPORT':
      raise DEVSException("In model '" + self.getModelFullName() + \
                          "', port '"+ port.getPortFullName() + \
                          "' is not an output port", 0)
     elif port.hostDEVS != self:
      raise DEVSException("Port '" + port.getPortFullName() + \
                          " does not belong to model '" + \
                          self.getModelFullName() + "'", 0)    
     else:
      # approved
      self.poke_ok[port] = True
    # The port itself is OK, but maybe we already poked it?
    if port in self.myOutput:
      raise DEVSException("In model '" + self.getModelFullName() + \
                          " an event has already been poked on port '" + \
                          port.getPortName() + "'", 0)
    else:
      # Doesn't seem to be the case, so do it!
      self.myOutput[port] = ev 
    
  ###
  def peek(self, port):
    """Retrieves message from input port {\tt port}.
    """
    
    # See if it is already checked
    if port not in self.peek_ok:
        if port.type() != 'INPORT':
           raise DEVSException("In model '" + self.getModelFullName() + \
                               "', port '"+ port.getPortName() + \
                               "' is not an input port", 0)
        elif port.hostDEVS != self:
           raise DEVSException("Port '" + port.getPortFullName() + \
                               " does not belong to model '" + \
                               self.getModelFullName() + "'", 0)
        else:
           # Checked and approved!
           self.peek_ok[port] = True
    # port certainly exists, just return whatever is there
    # Should nothing be there, return None like previously
    return self.myInput.get(port, None)

  ###    
  def extTransition(self):
    """DEFAULT External Transition Function.
  
    Accesses {\tt state} and {\tt elapsed} attributes, as well as inputs
    through {\tt peek} method. Returns the new state.
    """
    return self.state
    
  ###    
  def intTransition(self):
    """DEFAULT Internal Transition Function.
  
    Accesses only {\tt state} attribute. Returns the new state.
    """
    return self.state
  
  ###    
  def outputFnc(self):
    """DEFAULT Output Function.
  
    Accesses only {\tt state} attribute. Modify the output ports by means of
    the {\tt poke} method. Returns Nothing.
    """
    pass
  
  ###    
  def timeAdvance(self):
    """DEFAULT Time Advance Function.
  
    Accesses only the {\tt state} attribute. Returns a real number in
    $[0, +\infty]$.
    """
    # By default, return infinity 
    return float('inf')

    # MODIFIED 04/04/05: INFINITY stands for infinity
    
#    ===================================================================    #

class CoupledDEVS(BaseDEVS):
  """Abstract base class for all coupled-DEVS descriptive classes.
    
  A coupled-DEVS is described as a structure... (to be completed. Note that
  {\tt self} in coupled-DEVS' definition should not to be mistaken with Python's
  {\tt self})
  """
  
  # {\tt CoupledIDCounter}: See comment for {\tt AtomicIDCounter} above.
  CoupledIDCounter = 0
  
  ###  
  def __init__(self, name=None):
    """Constructor.
    """

    # Prevent any attempt to instantiate this abstract class
    if self.__class__ == CoupledDEVS:
      raise DEVSException("Cannot instantiate abstract class '%s' ... " % (self.__class__.__name__), 0)
    # The minimal constructor shall {\sl first\/} call the superclass
    # ({\it i.e.}, {\tt BaseDEVS}') constructor.
    BaseDEVS.__init__(self, name)
    
    # Increment {\tt CoupledIDCounter\/} and setup instance's {\tt myID\/}
    # attribute.
    CoupledDEVS.CoupledIDCounter = CoupledDEVS.CoupledIDCounter + 1
    self.myID = "C%d" % CoupledDEVS.CoupledIDCounter

    # {\sl Specific Attributes\/} --- related to coupled-DEVS'
    # specification: (to be completed)
    self.componentSet = []
    
    # {\tt IC}, {\tt EIC} and {\tt EOC} describe the couplings at the
    # coupled-DEVS level (respectively, {\sl internal couplings}, {\sl
    # external input couplings\/} and {\sl external output couplings\/}).
    # Each coupling in the sets are pairs of pairs referencing the component
    # and the port the coupling is beginning and the component and the port
    # the coupling is ending.
    # Note that although consistent with Zeigler's definition, these sets
    # are not used by the simulator, which relies on ports' {\tt inLine} and
    # {\tt outLine} attributes to detect couplings.
    self.IC  = []
    self.EIC = []
    self.EOC = []  

  ###
  def setHierarchy(self, hierarchy):
        # Add your own ID to the end of the list, since we are getting deeper
        hierarchy.append(self.auto_id)
        self.hierarchy = hierarchy
        for i in self.componentSet:
                newHierarchy = list(hierarchy)
                i.setHierarchy(newHierarchy)
      
  ###      
  def addSubModel(self, model):
    """Make {\tt model} a child of the coupled-DEVS (in the hierarchical
    representation of the model).
    
    This is the {\sl only\/} proper way to build a hierarchical model. Calls
    to {\tt addSubModel} can appear in coupled-DEVS' descriptive classes'
    constructors, or the method can be used later to modify an existing
    object. {\tt model} is an instance of an atomic- or coupled-DEVS'
    descriptive class.
    
    (to be completed: already instantiated in global space [can then be
    a whole model---one should then be careful multiple occurences of the
    same object] or instantiated when method is called [can then lead to
    recursion---one should be careful to avoid recursive definitions] )
    
    The method adds a reference to the sub-model in its parent's {\sl
    components set} $\{M_d\,\mid\,d\in{\cal D}\}$ and sets the sub-model's
    {\tt parent} attribute. The set of {\sl component references\/} $\cal D$
    is implicitly defined provided the modeler adds the returned reference
    the local dictionary.
    """
    self.componentSet.append(model)    
    model.parent = self
    return model
   
  def getSubModel(self, name):
    """Get a sub-model by name, return None if not found.
    """
    # HV: TODO: allow complete path names (e.g., "C1.C2.A") as names

    subModels = self.componentSet

    for i in subModels:
        if i.getModelName() == name:
                return i

    return None

  ###
  def connectPorts(self, p1, p2):
    """Connects two ports together. The coupling is to begin at {\tt p1} and
    to end at {\tt p2}.
    
    NOTE: connections should eventually be implemented as objects, which
    would allow to specify an input-output transformation.
    """

    # For a coupling to be valid, two requirements must be met:
    # 1- at least one of the DEVS the ports belong to is a child of the
    #    coupled-DEVS ({\it i.e.}, {\tt self}), while the other is either the
    #    coupled-DEVS itself or {\sl another\/} of its children. The DEVS'
    #   'parenthood relationship' uniquely determine the type of coupling;
    # 2- the types of the ports are consistent with the 'parenthood' of the
    #    associated DEVS. This validates the coupling determined above.
    
    # {\sl Internal Coupling\/}:
    if ((p1.hostDEVS.parent == self and p2.hostDEVS.parent == self) and
        (p1.type() == 'OUTPORT' and p2.type() == 'INPORT')):
      if p1.hostDEVS == p2.hostDEVS:
        raise DEVSException("In coupled model '" + self.getModelFullName() + \
                            "' connecting ports '" + p1.getPortFullName() + \
                            "' and '" + p2.getPortFullName() + \
                            "' belong to the same model '" + \
                            p1.hostDEVS.getModelFullName() + "'. " + \
                            "Direct feedback coupling not allowed", 0)
      else:
        self.IC.append( ( (p1.hostDEVS, p1), (p2.hostDEVS, p2) ) )
        p1.outLine.append(p2)
        p2.inLine.append(p1)
    
    # {\sl external input couplings\/}:
    elif ((p1.hostDEVS == self and p2.hostDEVS.parent == self) and
         (p1.type() == p2.type() == 'INPORT')):
        self.EIC.append( ( (p1.hostDEVS, p1), (p2.hostDEVS, p2) ) )
        p1.outLine.append(p2)
        p2.inLine.append(p1)
    
    # {\sl external output couplings\/}:
    elif ((p1.hostDEVS.parent == self and p2.hostDEVS == self) and
                (p1.type() == p2.type() == 'OUTPORT')):
        self.EOC.append( ( (p1.hostDEVS, p1), (p2.hostDEVS, p2) ) )
        p1.outLine.append(p2)
        p2.inLine.append(p1)

    # Other cases (illegal coupling):
    else:
      raise DEVSException("Illegal coupling in coupled model '" + \
                          self.getModelFullName() + "' between ports '" + \
                          p1.getPortFullName() + "' and '" + \
                          p2.getPortFullName() + "'", 0)
      
  ###    
  def select(self, immList):
    """DEFAULT Select Function.
  
    Take as a parameter a list of imminent children (DEVS) lexicographically
    sorted according to their {\tt myID} attribute. Returns an item from that
    list.
    """
    return immList[0]

##  PORT CLASS
## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##

class Port:
  """Class for DEVS model ports (both input and output).
  
  This class provides basic port attributes and query methods.
  """

  # {\tt InCounter} and {\tt OutCounter} are static counters used for the
  # 'index' part of input and output ports {\tt myID} attribute. Incremented
  # with every instantiation of input or output port, unaffected 
  # (i.e., no recycling of unique port numbers) by the
  # deletion of those ports (deletion not implemented yet).
  InCounter = 0 
  OutCounter = 0

  ###
  def __init__(self, isInput = True, name=None):
    """Constructor. Creates an input port if {\tt isInput} evaluates to True, and
    an output port otherwise.
    """
    
    # {\sl Port Attributes\/} (to be completed):
    # * {\tt inLine} and {\tt outLine} represent an alternate way to
    #   describe couplings (and the one actually used by the simulator). The
    #   former is a list of references to the (possibly many) ports from
    #   which {\sl this\/} port receives its messages, while the latter is a
    #   list of references to the (possibly many) ports which receive
    #   messages from {\sl this\/} port. (Note that no more than one message
    #   can be received at one time: to be completed). While an atomic-DEVS output
    #   port (input port) would only need to declare {\tt outLine}
    #   ({\tt inLine}), the dual nature of coupled-DEVS ports require both
    #   attributes;
    # * {\tt myID} (set below);
    # * {\tt hostDEVS} (to be completed);
    self.inLine = [] 
    self.outLine = []
    self.hostDEVS = None 
   
    # The name of the port
    self.name = name

    # Increment {\tt InCounter\/} or {\tt OutCounter\/} depending on type of
    # port desired, and setup instance's {\tt myID\/} attribute:
    if isInput:
      Port.InCounter += 1
      self.myID = "InPort_%d" % Port.InCounter
    else:
      Port.OutCounter +=  1
      self.myID = "OutPort_%d" % Port.OutCounter
        
  ###  
  def getPortName(self):
    if self.name is None:
      self.name = "port%s" % self.myID
      if self.hostDEVS is not None:
        attr_name = self.hostDEVS.getPortName(self)
        if attr_name is not None:
          self.name = attr_name
    return self.name

  def getPortFullName(self):
    base = ""
    if self.hostDEVS is not None:
      base = self.hostDEVS.getModelFullName() 
    return base + "." + self.getPortName()

  ###  
  def type(self):
    """Returns the 'type' of the object: {\tt 'INPORT'} or {\tt 'OUTPORT'}.
    """
    if self.myID[:2] == 'In':
      return 'INPORT'
    elif self.myID[:3] == 'Out':
      return 'OUTPORT'
    # Else, invalid type:
    return None


