# -*- coding: Latin-1 -*-
## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
## ## ##
# queueExperiments.py --- simple queueing system example
#                       --------------------------------
#                              October 2002
#                            Jean-Sebastien Bolduc
#                             Hans Vangheluwe 
#                         McGill University (Montréal)
#                       --------------------------------
## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
## ## ##

# Add the directory where pydevs lives to Python's import path
import sys
import os.path
sys.path.append(os.path.expanduser('~/UA/1MINF/OS1/pydevs-mod/'))

# Import code for model simulation:
import pydevs
from pydevs.infinity import *
from pydevs.simulator import *

#    ======================================================================    #

# Import the queueing system model to be simulated
from queueModels import *

#    ======================================================================    #
#  
#  The experiment
#

print "\n\n ** FIRST EXPERIMENT ******************************\n\n"

# 1. Instantiate the (Coupled or Atomic) DEVS at the root of the 
#  hierarchical model. This effectively instantiates the whole model 
#  thanks to the recursion in the DEVS model constructors (__init__).
#

#
# Reminder: the Root model has parameters 
#
#  np: number of processors
#  qs: queue size
#  ia: Job inter-arrival time lower bound
#  ib: Job inter-arrival time upper bound
#  sa: Job size lower bound
#  sb: Job size upper bound
#

# If the queue size is INFINITY (infinite capacity queue),
# only the first Processor will ever be used and Jobs will 
# never be DISCARDed.
#
theSystem = Root(np=2, qs=INFINITY, ia=3, ib=3, sa=4,  sb=6, name="theSystem")

# A more interesting case: queue size = 1
#
#theSystem = Root(np=2, qs=1, ia=2, ib=2, sa=7,  sb=7, name="theSystem")


# 2. Link the model to a DEVS Simulator: 
#  i.e., create an instance of the 'Simulator' class,
#  using the model as a parameter.
sim = Simulator(theSystem)


# 3. Run the simulator on the model until a termination_condition is met.
#  sim.simulate(termination_condition= ..., verbose= ...)
#

# Some typical termination_condition functions
#

# A. Simulate forever. 
#    The termination_condition function never returns True.
#
def terminate_never(model, clock):
  return False

#sim.simulate(termination_condition=terminate_never, 
#             verbose=True)

# B. Simulate until a specified end_time is reached.
#    When the end_time is reached/exceeded, the 
#    termination_condition function returns True.
#
def terminate_whenEndTimeReached(model, clock, end_time=1000):
  if clock >= end_time:
    return True
  else:
    return False

# Note how for the following simulation experiment,
# verbose=Falso, so no trace output will be sent to stdout
# (only the statistics gathered at the end of the experiment
# will be printed below).
#
sim.simulate(termination_condition=terminate_whenEndTimeReached, 
             verbose=False)

# To print the full trace instead:
#
#sim.simulate(termination_condition=terminate_whenEndTimeReached, 
#             verbose=True)

# C. Simulate until a specified end_state of one of the sub-models is reached.
#    When the end_stateis reached, the 
#    termination_condition function returns True.
#
def terminate_whenStateIsReached(model, clock, end_state=False):
  if model.GDEVS.state.first == end_state:
    return True
  else:
    return False

#sim.simulate(termination_condition=terminate_whenStateIsReached, 
#             verbose=False)

print "Statistics:\n"

firstProcessor = theSystem.getSubModel(name="Processor0")

print "  Number of Jobs processed by Processor0 = %s" % firstProcessor.state.numberOfJobsProcessed

averageProcessingTime = firstProcessor.state.totalProcessingTime / \
                        firstProcessor.state.numberOfJobsProcessed
print "  Average job processing time by Processor0 = %f\n" % averageProcessingTime


print "\n\n ** SECOND EXPERIMENT ******************************\n\n"

# Re-set the global Job counter
#
Job.IDCounter = 0

# 1. Instantiate the (Coupled or Atomic) DEVS at the root of the 
#  hierarchical model. This effectively instantiates the whole model 
#  thanks to the recursion in the DEVS model constructors (__init__).
#
theSystem = Root2(name="theSystem")

# 2. Link the model to a DEVS Simulator: 
#  i.e., create an instance of the 'Simulator' class,
#  using the model as a parameter.
sim = Simulator(theSystem)

# 3. Run the simulator on the model until a termination_condition is met.
#  sim.simulate(termination_condition= ..., verbose= ...)
#

#  Simulate until a specified end_time is reached.
#  When the end_time is reached/exceeded, the 
#  termination_condition function returns True.
#
def terminate_whenEndTimeReached(model, clock, end_time=20):
  if clock >= end_time:
    return True
  else:
    return False

# Note how for the following simulation experiment,
# verbose=Falso, so no output will be sent to stdout.
#
sim.simulate(termination_condition=terminate_whenEndTimeReached, dconnect=False,
             verbose=True)

print "\n\n **************************************************\n\n"
