# -*- coding: Latin-1 -*-
## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
## ## ##
# queueModels.py --- simple queueing system examples
#                       --------------------------------
#                              October 2002
#                            Jean-Sebastien Bolduc
#                             Hans Vangheluwe 
#                         McGill University (Montréal)
#                       --------------------------------
## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
## ## ##

# Add the directory where pydevs lives to Python's import path
import sys
import os.path
sys.path.append(os.path.expanduser('~/UA/1MINF/OS1/pydevs-mod/'))

# Import code for model representation:
import pydevs
from pydevs.devs_exceptions import *
from pydevs.infinity import *
from pydevs.DEVS import *

# Import for uniform random number generators
from random import uniform
from random import randint

#    ======================================================================    #

class Job:
  """ A job has a unique, strictly positive integer ID number, 
      and a positive 'size', in this case directly indicating 
      the time a processor needs to process this job.
  """
  IDCounter = 0   # class variable, globally unique

  def __init__(self, szl, szh):
    """ The job size is drawn from the interval [szl, szh[ 
        (uniform distribution)
    """
    Job.IDCounter += 1
    self.ID = Job.IDCounter

    self.size = uniform(szl, szh)

  def __str__(self):
    return "(job %d, size %f)" % (self.ID, self.size)

#    ======================================================================    #

class GeneratorState:
  """ A class to hold the state of a Generator Atomic DEVS.
  
  Not really required, but allows us to define 
  the __str__ method for pretty printing.
  """

  def __init__(self, first):
    # keep track whether this is the first
    # time an internal transition is invoked
    self.first = first 

  def __str__(self):
    return "\n\
            first           = %s" % self.first

class Generator(AtomicDEVS):
  """ Generates jobs.

  The inter arrival time between job is an integer randomly chosen from the
  interval [ia, ib[ (uniform distribution).

  The size of a job is an integer randomly chosen from the
  interval [sa, sb[ (uniform distribution). 

  We want to be able to generate the first job at time 0.
  As output is only produced at internal transition times,
  so we need an internal transition at time 0.

  We could initialize the self.elapsed time to the value
  returned by the very first invocation of timeAdvance.
  This would force an internal event at time 0.
  This alternative is commented out below.

  The other possibility is to schedule a new event to occur
  after 0 time. For this, timeAdvance needs to return 0 the
  first time it is invoked.
  """

  # all arguments should be strictly positive integers
  def __init__(self, ia, ib, sa, sb, name=None):

    # Always call parent class' constructor FIRST:
    AtomicDEVS.__init__(self, name) 
   
    # make a local copy of the constructor's arguments
    self.ia = ia 
    self.ib = ib
    self.sa = sa 
    self.sb = sb
   
    # add one output port
    self.OUT = self.addOutPort(name="OUT")

    self.state=GeneratorState(first=True)
    # Alternative:
    # self.elapsed = arbitrary_number
    
  def intTransition(self):
    self.state.first = False 
    return self.state
    
  def outputFnc(self):
    job = Job(self.sa, self.sb)        
    # In pythonDEVS, event objects are "poke"d on output ports
    # in the outputFnc. Conversely, in the extTransition
    # function, it is possible to "peek" input ports for incoming
    # event objects. This deviates from classic DEVS which 
    # does not have ports.
    self.poke(self.OUT, job)

  def timeAdvance(self):
    if self.state.first == True:
     return 0
    else:
     return uniform(self.ia, self.ib)
    # Alternative
    # if self.state.first == True:
    #  return arbitrary_number 
    # else:
    #  return uniform(self.ia, self.ib)
    # 
    
#    ======================================================================    #

# Some useful literals
# for processorStatus
IDLE     = 0
BUSY     = 1
DISCARD  = 2

class ProcessorState:
  """ A class to hold the state of a Processor Atomic DEVS.
  
  Not really required, but allows us to define 
  the __str__ method for pretty printing.
  """

  def __init__(self, queue=[], 
                     processorStatus=IDLE, 
                     currentJob=None, 
                     timeSpent=0):
    self.queue = queue
    self.processorStatus = processorStatus 
    self.currentJob = currentJob # None if the processor is idle
    self.timeSpent = timeSpent
    # timeSpent is the time spent so far on the current process. This is
    # necessary as the DEVS elapsed time is reset after an external event
    # has occured (i.e., when a new job is received from the generator) !
    
    # To be able to calculate average processing time performance
    # metric, need to add extra state variables.
    self.currentTime = 0           # as seen by this Processor
    self.processingStartTime = 0   # of current Job
    self.processingEndTime = 0     # of current Job
    self.totalProcessingTime = 0   # of all Jobs, by this Processor
    self.numberOfJobsProcessed = 0 # by this Processor

  def __str__(self):
    str_repr = "\n"
    if len(self.queue) == 0:
      str_repr +=  "            queue           = []\n"
    else: 
      str_repr +=  "            queue           = [\n"
      for i in range(len(self.queue)-1):
        str_repr +=  "                               %s,\n" % self.queue[i]
      str_repr +=  "                               %s\n" % self.queue[-1]
      str_repr +=  "                              ]\n"
    str_repr +=  "            processorStatus = %s\n" \
      % ("IDLE", "BUSY", "DISCARD")[self.processorStatus]
    str_repr +=  "            currentJob      = %s\n" % self.currentJob
    str_repr +=  "            timeSpent       = %s\n" % self.timeSpent
    str_repr += "\n"

    # Note: we don't print the extra state variables used
    # to calculate performance measures (average processing time
    # in this case).

    return str_repr

class Processor(AtomicDEVS):
  """Processes jobs.
  
  The processor has a queue of size queue_size (default: INFINITY,
  which means the queue is not bounded) to store jobs.
  Jobs are processed for a time equal to their size attribute.
  When a job is finished, it is sent through the OUT output port.
  If the queue is full, incoming jobs are sent through the DISCARD output port.
  """
  
  def __init__(self, queue_size=INFINITY, name=None):
    AtomicDEVS.__init__(self, name)    

    # make a local copy of the constructor's argument
    # n is a parameter
    self.queue_size = queue_size 

    # atomic-DEVS' state: 
    self.state = ProcessorState(queue=[],
                                processorStatus=IDLE,
                                currentJob=None,
                                timeSpent=0)

    # Declare ports
    self.IN      = self.addInPort("IN")
    self.OUT     = self.addOutPort("OUT")
    self.DISCARD = self.addOutPort("DISCARD")

  def extTransition(self):

    # For statistics gathering
    self.state.currentTime += self.elapsed 
    #print "in external: current time = %s\n" % self.state.currentTime

    # Enqueue incoming message ("virtually", if the queue is full)
    job = self.peek(self.IN)
    self.state.queue.append(job) 

    # If IDLE, make the processor BUSY with the job at the head of the queue.
    if self.state.processorStatus == IDLE:
      self.state.processorStatus = BUSY 
      self.state.currentJob = self.state.queue[0]
      # Don't forget to trim the queue.
      self.state.queue = self.state.queue[1:]

      # gather statistics
      self.state.processingStartTime = self.state.currentTime

    # If BUSY, update timeSpent on job (to re-schedule end of Job):
    elif self.state.processorStatus == BUSY:
      self.state.timeSpent += self.elapsed
      if len(self.state.queue) > self.queue_size:
        self.state.processorStatus = DISCARD 

    # If DISCARD, something went wrong as we should not
    # receive external interrupts while in DISCARD mode:
    else:
      raise DEVSException(\
            "should not receive external interrupts while in DISCARD mode")

    return self.state

  def intTransition(self):

    # For statistics gathering
    # BE CAREFUL ! This only works if timeAdvance is deterministic !!!
    # Otherwise, need to keep one extra state variable
    # to remember what the time advance was.
    self.state.currentTime += self.timeAdvance() 
    #print "in internal: current time = %s\n" % self.state.currentTime

    if self.state.processorStatus == BUSY:

      # For statistics gathering
      self.state.processingEndTime = self.state.currentTime  # of finished Job
      processingTime = \
        self.state.processingEndTime - self.state.processingStartTime
      self.state.totalProcessingTime += processingTime 
      self.state.numberOfJobsProcessed += 1 
      #print "Job leaves at time %s\n" % self.state.currentTime
      #print "processingTime = %s\n" % processingTime
      #print "numberOfJobsProcessed = %s\n" % self.state.numberOfJobsProcessed

      if len(self.state.queue) > 0:
      # If was BUSY and the queue is not empty, the processor remains BUSY
      # but with a new Job.
        self.state.currentJob = self.state.queue[0]
        # don't forget to trim the queue.
        self.state.queue = self.state.queue[1:]
        
        # For statistics gathering
        self.state.processingStartTime = self.state.currentTime  # of new Job

      else:
      # If was BUSY but the queue is empty, becomes IDLE
        self.state.processorStatus = IDLE 
        self.state.currentJob = None 

      self.state.timeSpent = 0

    elif self.state.processorStatus == DISCARD:
    # If the processor was about to DISCARD a job,
      self.state.queue = self.state.queue[:-1]
      if len(self.state.queue) <= self.queue_size:
        # if the queue length is again <= queue_size: switch to BUSY mode
        self.state.processorStatus = BUSY 

        # For statistics gathering
        self.state.processingStartTime = self.state.currentTime  # of new Job

      # if the queue length is still too large: remain in DISCARD mode
  
    # No internal transition should happen when the processor is IDLE,
    # since the time advance is then infinity.
    else:
      raise DEVSException(\
            "no internal transition should occur when the processor is IDLE\n", 
            "since the time advance is then INFINITY")
    
    return self.state

  def outputFnc(self):
    if self.state.processorStatus == BUSY:
      # Put the processed message on the output port:
      self.poke(self.OUT, self.state.currentJob)
    elif self.state.processorStatus == DISCARD:
      self.poke(self.DISCARD, self.state.queue[-1])
    else:
      raise DEVSException(\
            "no output should be generated when in IDLE mode")
    
  def timeAdvance(self):
    # Return the remaining processing time for current process if any (BUSY),
    # or a large number that stands for infinity (IDLE).
    if self.state.processorStatus == BUSY:   
      # re-schedule termination of job processing
      return self.state.currentJob.size - self.state.timeSpent
    elif self.state.processorStatus == DISCARD: 
      # schedule output of discarded job at current time
      return 0
    else:
      # stay in IDLE state forever
      return INFINITY

#    ======================================================================    #
#    Cascaded Processors example

class Root(CoupledDEVS):

  def __init__(self, np = 1, qs = 5, ia = 2, ib = 7, sa = 5, sb = 10, name=None):
    """Constructor.
    """
  
    # Always call parent class' constructor FIRST:
    CoupledDEVS.__init__(self, name)
    
    # Declare the coupled model's output ports:
    self.OUT     = self.addOutPort("OUT")
    self.DISCARD = self.addOutPort("DISCARD")    

    # Declare the coupled model's sub-models:

    # The generator
    self.GDEVS = self.addSubModel(Generator(ia,ib,sa,sb, name="GeneratorDEVS"))
   
    # n cascaded processors
    prevPORT = self.GDEVS.OUT

    for i in range(0,np):
      tmpName = "Processor%s" % i
      tmp = self.addSubModel(Processor(qs, name=tmpName))
      self.connectPorts(prevPORT, tmp.IN)
      self.connectPorts(tmp.OUT, self.OUT) 
      prevPORT = tmp.DISCARD

    self.connectPorts(prevPORT, self.DISCARD) 	      

  def select(self, immList):
    """Choose the sub-model furthest down the cascade.
       This is reasonable as it lets downstream processors
       finish first, possibly allowing new jobs in.
    """
    return immList[-1]

    # Alternative: randomly choose among imminent submodels
    #return immList[randint(0,len(immList))]

    # If the select method is not defined, the immList[0] 
    # will be chosen by default.

##    ======================================================================    #
## An alternate Root model demonstrating collisions 
##
class Root2(CoupledDEVS):

  def __init__(self, name=None):
    CoupledDEVS.__init__(self, name)
    self.OUT = self.addOutPort(name="OUT")

    # Declare the two sub-models:
    self.Processor1 = \
      self.addSubModel(Processor(queue_size=0, name="Processor0"))
    self.Processor2 = \
      self.addSubModel(Processor(queue_size=0, name="Processor1"))
    
    self.connectPorts(self.Processor1.OUT, self.Processor2.IN)
    self.connectPorts(self.Processor2.OUT, self.OUT)
    
    # Creates a loop:
    #self.connectPorts(self.Processor2.OUT, self.Processor1.IN)
    
    job1 = Job(1, 1)
    self.Processor1.state = ProcessorState(processorStatus=BUSY,
                                           currentJob=job1)
    job2 = Job(1, 1)
    self.Processor2.state = ProcessorState(processorStatus=BUSY,
                                           currentJob=job2)
    
    
  def select(self, immList): # Alternative select method
    return immList[-1]  

#    ======================================================================    #

