
## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
# templateModel.py --- Template for atomic- and coupled-DEVS descriptive classes
#                       --------------------------------
#                              Copyright (c) 2000
#                            Jean-Sbastien  BOLDUC
#                         McGill University (Montral)
#                       --------------------------------
#       created: 11/07/00
# last modified: 21/11/05 by Hans Vangheluwe
## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##

# Add the directory where pydevs lives to Python's import path
import sys
import os.path
sys.path.append(os.path.expanduser('~/src/projects/PythonDEVS/'))

# Import code for DEVS model representation:
import pydevs
from pydevs.devs_exceptions import *
from pydevs.infinity import *
from pydevs.DEVS import *

# Import for uniform random number generators 
from random import uniform
from random import randint

#    ======================================================================    #

class SampleEvent:

  """Models communicate (through ports) by sending/receiving events.
     These events (often used to encode a GPSS-style "transaction")
     can have structure. It is hence useful
     to use event objects which are instances of a class.

     With an appropriate __str__ method (giving a string representation
     of the class), these events will also be properly displayed 
     in a simulation trace.
  """
  def __init__(self, parameter= ...):
    """Constructor (parameterizable).
    """

    # Often we want a unique ID for each SampleEvent instance:
    # use a class variable incremented with each new instantiation.
    SampleEvent.uniqueID += 1

    # Set some Event parameters
    self.param1 = ... 
    # Often, these parameters are initialized with values
    # taken from constructor arguments 
    self.param2 = parameter 
    # Often, these parameters are drawn from a random distribution
    self.param3 = uniform(lower_bound, upper_bound)  # Real number
    self.param4 = randint(lower_bound, upper_bound)  # Integer number

  # string representation
  def __str__(self):
    return "(param1 %d, param2 %f, ...)" % (self.param1, self.param2, ...)

# Initialize the class variable
SampleEvent.uniqueID = 0 

#    ======================================================================    #

class SampleADEVSState:

  """Encapsulate the Atomic DEVS' state
     This is not absolutely necessary, but it makes
     it easier to 
     (1) stick to the DEVS formalism
     (and not modify the state "behind the back of the simulator" --
     this will be enforced once we have a DEVS compiler), and
     (2) to produce a clean simulation trace (thanks to __str__).
  """

  ###
  def __init__(self, parameters= ...):
    """Constructor (parameterizable).
    """

    # set the initial state
    # the state may be multi-dimensional: 
    # consist of multiple state variables
    self.stateVar1 = parameter ... 
    self.stateVar2 = ...

  # string representation
  def __str__(self):
    return a string representing the entire SampleADEVSState 

class SampleADEVS(AtomicDEVS):
  """Sample Atomic-DEVS descriptive class.
  """
  
  ###
  def __init__(self, some parameters= ..., name=None):
    """Constructor (parameterizable).
       Giving a name (string) to a model is not mandatory
       (a globally unique name "A<i>" is assigned by default).
       If a name is given, the simulation trace becomes far more
       readable.
    """
    
    # Always call parent class' constructor FIRST:
    AtomicDEVS.__init__(self, name=name)
    
    # TOTAL STATE:
    #  Define 'state' attribute (initial sate):
    self.state = SampleADEVSState(initial settings ...) 

    # ELAPSED TIME:
    #  Initialize 'elapsed time' attribute if required.
    #  (by default (if not set by user), self.elapsed is initialized to 0.0)
    self.elapsed = 1.1 
    #  With elapsed time for example initially 1.1 and 
    #  this SampleADEVS initially in 
    #  a state which has a time advance of 60,
    #  there are 60-1.1 = 58.9 time-units remaining until the first 
    #  internal transition 

    # HV: check in simulator that elapsed can not be larger than
    # timeAdvance
    
    # PORTS:
    #  Declare as many input and output ports as desired
    #  (usually store returned references in local variables):
    self.IN = self.addInPort(name="IN")
    self.OUT = self.addOutPort(name="OUT")

  ###
  def extTransition(self):
    """External Transition Function."""
    
    # Compute and return the new state based (typically) on current
    # State, Elapsed time parameters and calls to 'self.peek(self.IN)'.
    # When extTransition gets invoked, input has arrived
    # on at least one of the input ports. Thus, peek() will
    # return a non-None value for at least one input port.

    input = self.peek(self.IN)

    state = self.state

    # Typical logic:

    if input == some_input_1:
      if state == some_state_1:
       return SampleADEVSState(new state parameters)
      if state in (some set of states):
       return SampleADEVSState(other new state parameters)
      else:
       raise DEVSException(\
        "unknown state <%s> in SampelADEVS external transition function"\
        % state)
    
    if input == some_input_2:
      ...

    ...

    raise DEVSException(\
      "unknown input <%s> in SampelADEVS external transition function"\
      % input) 

  ###
  def intTransition(self):
    """Internal Transition Function.
    """

    # Compute and return the new state  based (typically) on current State.
    # Remember that intTransition gets called after the outputFnc,
    # timeAdvance time after the current state was entered.
  
    state = self.state

    # Typical logic:

    if state == some_state_1:
      return SampleADEVSState(new state parameters)
    if state in (some set of states):
      return SampleADEVSState(other new state parameters)
    ...
    else:
      raise DEVSException(\
      "unknown state <%s> in SampelADEVS external transition function"\
      % state)

  ###
  def outputFnc(self):
    """Output Function.
    """
    
    # Send events via a subset of the atomic-DEVS' 
    # output ports by means of the 'poke' method.
    # More than one port may be poke-d.
    # The content of the messages is based (typically) on the current State.
    # 
    # BEWARE: ouput is based on the CURRENT state
    # and is produced BEFORE making the internal transition.
    # Often, we want output based on the NEW state (which
    # the system will transition to by means of the internal 
    # transition. The solution is to keep the NEW state in mind
    # when poke-ing the output. 
    # This will only work if the internal transition function 
    # is deterministic !

    state = self.state

    if state == some_state_1:
      self.poke(self.OUT, SampleEvent(some params))
    elif state == some_state_2:
      self.poke(self.OUT, SampleEvent(some other params))
      # NOT self.poke(self.OBSERVED, "grey")
    ...
    else:
      raise DEVSException(\
        "unknown state <%s> in SampelADEVS external transition function"\
        % state)
    
  ###
  def timeAdvance(self):
    """Time-Advance Function.
    """
    
    # Compute 'ta', the time to the next scheduled internal transition,
    # based (typically) on the current State.
    
    # Note how it 
    state = self.state

    if state == some_state_1:
      return 20
    elif state == some_state_2:
      return 10.123 
    ...
    elif state == passive_state:
      return INFINITY 
    else:
      raise DEVSException(\
        "unknown state <%s> in SampelADEVS time advance function"\
        % state)

    return ta

#    ======================================================================    #

class SampleCDEVS(CoupledDEVS):
  """Sample Coupled-DEVS descriptive class: compulsory material.
  """
  
  ###
  def __init__(self, name=None):
    """Constructor (parameterizable).
       Giving a name (string) to a model is not mandatory
       (a globally unique name "C<i>" is assigned by default).
       If a name is given, the simulation trace becomes far more
       readable.
    """
  
    # Always call parent class' constructor FIRST:
    CoupledDEVS.__init__(self, name)
    
    # PORTS:
    #  Declare as many input and output ports as desired
    #  (usually store returned references in local variables).
    #  Note that giving a name (string) to a port is not mandatory
    #  (a globally unique name "port<i>" is assigned by default).
    #  If a name is given, the simulation trace becomes far more
    #  readable however.
    self.IN = self.addInPort(name="IN")
    self.OUT = self.addOutPort(name="OUT")
    #  Note: at the root level, the Coupled DEVS will often
    #  not have any InPorts nor OutPorts (i.e., be "autonomous").

    # SUB-MODELS:
    #  Declare as many sub-models as desired. Method 'addSubModel' takes as a
    #  parameter an INSTANCE of an atomic- or coupled-DEVS descriptive class,
    #  and returns a reference to that instance (usually kept in a local
    #  variable for future reference):
    self.SUB = self.addSubModel(SampleADEVS(name="SUB")) 
    
    # COUPLINGS:
    #  Method 'connectPorts' takes two parameters 'p1' and 'p2' which are
    #  references to ports. 
    #  The coupling is to BEGIN at 'p1' and to END at 'p2'.
    #  Each input/output port can have 
    #  multiple incoming/outgoing connections.
    #  Note how PythonDEVS does NOT support Z_i,j transfer
    #  functions (as defined in DEVS) yet.
    #  On the other hand, PythonDEVS does have named ports.
    self.connectPorts(self.SUB1.OUT1, self.SUB2.IN1) # INTERNAL COUPLING
    self.connectPorts(self.IN, self.SUB.IN)   # EXTERNAL INPUT COUPLING
    self.connectPorts(self.SUB.OUT, self.OUT) # EXTERNAL OUTPUT COUPLING
      
  ###
  def select(self, immList):
    """Select function.
    
    Parameter 'immList' is a list of references to sub-models which
    have a scheduled internal transition occuring at the same time.
    To implement tie-breaking, select() needs to return 
    exactly ONE submodel immList[i].
    """

    # Several approaches are common:

    # 1. if the select method is not defined, the immList[0] 
    # will be chosen by default.
    return immList[0]

    # 2. the dual: chose the last entry
    return immList[-1]

    # 3. randomly choose among imminent submodels
    return immList[randint(0,len(immList))]

    # 4. A common technique is to choose the sub-model
    #    with the highest priority.

    # return the first instance of class CL in the immList
    for i in range(len(immList)):
      if immList[i].__class__ == CL:
        return immList[i]
    # if no CL instances found, return the last entry
    return immList[-1]

    #    To select between sub-models of the same type, 
    #    the name or unique ID should be used.

#    ======================================================================    #

