# -*- coding: Latin-1 -*-
## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
# simulator.py --- 'Plain' DEVS Model Simulator
#                     --------------------------------
#                            Copyright (c) 2000
#                          Jean-Sébastien  BOLDUC
#                             Hans  Vangheluwe
#                            Yentl Van Tendeloo
#                       McGill University (Montréal)
#                     --------------------------------
# Version 1.0                                        last modified: 01/11/01
# Version 1.0.1                                      last modified: 04/04/05
#  - Deals with DeprecationWarning
# Version 1.0.2                                      last modified: 06/06/05
#  - Added code to display atomic DEVS' initial conditions
# Version 1.1                                        last modified: 09/09/12
#  - A lot of optimizations
#    
## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
#
# GENERAL NOTES AND REMARKS:
#
# Atomic- and coupled-DEVS specifications adapted from:
#       B.P.Zeigler, H. Praehofer, and Tag Gon Kim,
#       ``Theory of modeling and simulation: Integrating
#       Discrete Event and Continuous Complex Dynamic Systems '',
#       Academic Press, 2000
#
# Description of message passing mechanism (to be completed)
#
## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##


##  IMPORTS 
## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##

# Import for uniform random number generators
import random

from heapq import *

# Import generic stuff 
from devs_exceptions import *
from math import floor

import functools

from VCDRecord import *

# Necessary to send COPIES of messages.
from copy import deepcopy

##  GLOBAL VARIABLES AND FUNCTIONS
## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##

# {\tt __VERBOSE__} switch allows to enter\dots well, ''verbose mode'' when set
# to true. Useful for debugging (similar to {\tt #define _DEBUG_} convention in
# {\tt C++}).
# __VERBOSE__ = True
# Now a global variable set by an argument passed to the simulator

# Tolerance for checking (in)equality of real numbers
EPSILON = 1E-10

id_fetch = []
__DIRECTCONNECT__ = True

# Old timestamp for VCD export
PREVIOUSTIMESTAMP = 0

## Print functions
##############################################################################

def printCollisionInfo(cDEVS, immChildren, dStar):
        if cDEVS.name == None:
          print "\n\tCOLLISION occured in model <%s> involving:" % (cDEVS.myID)
        else:
          print "\n\tCOLLISION occured in model <%s> involving:" % (cDEVS.name)
        for I in immChildren:
          if I.name == None:
            print "    \t   sub-model <%s>" % (I.myID)
          else:
            print "    \t   sub-model <%s>" % (I.name)
        if dStar.name == None:
          print "\tselect chooses sub-model <%s>" % (dStar.myID)
        else:
          print "\tselect chooses sub-model <%s>" % (dStar.name)

def printInitialCondition(model):
      ### ADDED 06/06/05: code added to display initial conditions:      
      if __VERBOSE__:
        if model.type() == 'COUPLED':
                # A coupled model doesn't have a initial state
                return
        print "\n\tINITIAL CONDITIONS in model <%s>" % model.getModelFullName()
        print "\t  Initial State: %s" % str(model.state)
        print "\t  Next scheduled internal transition at time %s" % `model.timeNext`


## trace processing functions
##############################################################################

def saveTrace(modelName, timeStamp, eventKind, portInfo,xmlState, strState):
        traceList.append( "<event>\n"
                  + "<model>"+modelName+"</model>\n"
                  + "<time>"+timeStamp+"</time>\n"
                        + "<kind>"+eventKind+"</kind>\n"
                  + portInfo
                        + "<state>\n"+xmlState+"<![CDATA["+strState+"]]>\n"+"</state>\n"
                        + "</event>\n"
                  )     

def saveVCD(modelName, portName, vcdState):
        # Check if the signal is a valid binary signal
        for i in range(len(vcdState)):
                if (i == 0):
                        if vcdState[i] == 'b':
                                continue
                        else:
                                raise DEVSException("Port " + portName + " in model " + modelName + " does not carry a binary signal\nVCD exports should carry a binary signal")
                char = vcdState[i]
                if char not in ["0", "1", "E", "x"]:
                        raise DEVSException("Port " + portName + " in model " + modelName + " does not carry a binary signal\nVCD exports should carry a binary signal")
        # Find the identifier of this wire
        for i in range(len(varList)):
                if varList[i].modelName == modelName and varList[i].portName == portName:
                        identifier = str(varList[i].identifier)
                        recordNr = i
                        break
        # If the bitSize is not yet defined, define it now
        if varList[i].bitSize == None:
                varList[i].bitSize = len(vcdState)-1
        elif varList[i].bitSize != len(vcdState) - 1:
                raise DEVSException("Wire has changing bitsize!\nYou are probably not using bit encoding on the wires!")
        # Now we have to convert between logisim and VCD notation
        vcdState = vcdState.replace('x', 'z')
        vcdState = vcdState.replace('E', 'x')
        # identifier will be defined, otherwise the record was not in the list
        vcdList.append(vcdState + " " + identifier)


def vcdHeader():
        string = ""
        string += "$date\n"
        from datetime import date
        string += "\t" + date.today().isoformat() + "\n"
        string += "$end\n"
        string += "$version\n"
        string += "\tPyDEVS VCD export\n"
        string += "$end\n"
        string += "$comment\n"
        string += "\tGenerated from DEVS-code\n"
        string += "$end\n"
        string += "$timescale 1ns $end\n"
        return string

def vcdVariables():
        modelList = []
        for i in range(len(varList)):
                if varList[i].modelName not in modelList:
                        modelList.append(varList[i].modelName)
        string = ""
        for module in modelList:
                string += "$scope " + module + " " + module + " $end\n"
                for var in range(len(varList)):
                        if varList[var].modelName == module:
                                string += "$var wire " 
                                if varList[var].bitSize == None:
                                        string += "1"
                                else:
                                        string += str(varList[var].bitSize)
                                string += " " + str(varList[var].identifier) + " " + varList[var].portName + " $end\n"
                string += "$upscope $end\n"
        string += "$enddefinitions $end\n"
        string += "$dumpvars \n"
        for var in range(len(varList)):
                string += "b"
                if varList[var].bitSize == None:
                        # The wire is a constant error signal, so the wire is never used
                        # Assume 1 bit long
                        string += "z"
                else:
                        for i in range(varList[var].bitSize):
                                string += "z"
                string += " " + str(varList[var].identifier) + "\n"
        string += "$end\n"
        return string

##  SIMULATOR CLASSES
## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##

class AtomicSolver:
  """Simulator for atomic-DEVS.
  
  Atomic-DEVS can receive three types of messages: $(i,\,t)$, $(x,\,t)$ and
  $(*,\,t)$. The latter is the only one that triggers another message-
  sending, namely, $(y,\,t)$ is sent back to the parent coupled-DEVS. This is
  actually implemented as a return value (to be completed).
  """
  
  ###
  def receive(self, aDEVS, msg):
    
    # For any received message, the time {\tt t} (time at which the message
    # is sent) is the second item in the list {\tt msg}.
    t = msg[1]
  
    # $(*,\,t)$ message --- triggers internal transition and returns
    # $(y,\,t)$ message for parent coupled-DEVS:
    if msg[0] == 1:
      if abs(t - aDEVS.timeNext) >= EPSILON:
        raise DEVSException("Bad synchronization in atomic model '" + \
                            aDEVS.getModelFullName() + "' on internal " + \
                            "transition; simulation time = " + str(t) + \
                            "; model.timeNext = " + str(aDEVS.timeNext), 1)

      # First call the output function, which (amongst other things) rebuilds
      # the output dictionary {\tt myOutput}:
      aDEVS.myOutput = {}
      aDEVS.elapsed = t - aDEVS.timeLast
      aDEVS.outputFnc()
                
      # Call the internal transition function to update the DEVS' state, and
      # update time variables:
      aDEVS.state = aDEVS.intTransition()      
      aDEVS.timeLast = t
      ta = aDEVS.timeAdvance()
      if ta < 0:
        raise DEVSException("Negative time advance in atomic model '" + \
                                aDEVS.getModelFullName() + "' with value " + \
                                str(ta) + " at time " + str(t))
      aDEVS.timeNext = aDEVS.timeLast + ta
      aDEVS.elapsed = 0.
      
      if __VERBOSE__:
#          print "\n\tINTERNAL TRANSITION in model <%s>" % (aDEVS.myID)
        print "\n\tINTERNAL TRANSITION in model <%s>" % aDEVS.getModelFullName()
        print "\t  New State: %s" % str(aDEVS.state)
        print "\t  Output Port Configuration:"
        for I in range(0, len(aDEVS.OPorts) ):
          if aDEVS.OPorts[I] in aDEVS.myOutput:
            print "\t    port <%s>: %s" % (aDEVS.OPorts[I].getPortName(),\
                                           aDEVS.myOutput[aDEVS.OPorts[I]])
#            print "\t    port%d: %s at %s" % (I, aDEVS.myOutput[aDEVS.OPorts[I]], \
#              repr(aDEVS.myOutput[aDEVS.OPorts[I]]))
          else:
            print "\t    port%d: NoEvent" % (I)
        print "\t  Next scheduled internal transition at time %s" % `aDEVS.timeNext`

      ### ADDED 01/09/2006
      if __TRACE__:
        portInfo = ""
        for I in range(0, len(aDEVS.OPorts) ):
          if aDEVS.OPorts[I] in aDEVS.myOutput:
            portInfo += "<port name=\""+ aDEVS.OPorts[I].getPortName()+"\" category=\"O\">\n"
            portInfo += "<message>%s</message>\n</port>\n" % aDEVS.myOutput[aDEVS.OPorts[I]]
        saveTrace(aDEVS.getModelFullName(), str(aDEVS.timeLast), "IN", portInfo, aDEVS.state.toXML(),str(aDEVS.state))
      
      ### ADDED 21/09/2011
      if __VCD__:
        for I in range(0, len(aDEVS.OPorts) ):
          if aDEVS.OPorts[I] in aDEVS.myOutput:
            portName = aDEVS.OPorts[I].getPortName()
            portSignal = aDEVS.myOutput[aDEVS.OPorts[I]]
            if portSignal != None:
                saveVCD(aDEVS.getModelFullName(), portName, str(portSignal))

      # Return the DEVS' output to the parent coupled-DEVS (rather than
      # sending $(y,\,t)$ message).
      return aDEVS.myOutput

    # $(x,\,t)$ message --- triggers external transition, where $x$ is the
    # input dictionary to the DEVS:
    elif type(msg[0]) == type({}):
      if not(aDEVS.timeLast - EPSILON <= t <= aDEVS.timeNext + EPSILON):
        raise DEVSException("Bad synchronization in atomic model '" + \
                            aDEVS.getModelFullName() + "' on external " + \
                            "transition; simulation time = " + str(t) + \
                            "; model.timeLast = " + str(aDEVS.timeLast) + \
                            "; model.timeNext = " + str(aDEVS.timeNext), 1)
      
      aDEVS.myInput = msg[0]
      # update elapsed time. This is necessary for the call to the external
      # transition function, which is used to update the DEVS' state.
      aDEVS.elapsed = t - aDEVS.timeLast
      aDEVS.state = aDEVS.extTransition()      
      # Udpate time variables:
      aDEVS.timeLast = t
      ta = aDEVS.timeAdvance()
      if ta < 0:
        raise DEVSException("Negative time advance in atomic model '" + \
                                aDEVS.getModelFullName() + "' with value " + \
                                str(ta) + " at time " + str(t))
      aDEVS.timeNext = aDEVS.timeLast + ta
      aDEVS.elapsed = 0
      
      if __VERBOSE__:
        print "\n\tEXTERNAL TRANSITION in model <%s>" % aDEVS.getModelFullName()
        print "\t  Input Port Configuration:"
        for I in range(0, len(aDEVS.IPorts) ):
          print "\t    port <%s>: %s" % (aDEVS.IPorts[I].getPortName(),\
                                         aDEVS.peek(aDEVS.IPorts[I]))
#          print "\t    port%d: %s at %s" % (I, aDEVS.peek(aDEVS.IPorts[I]),\
#            repr(aDEVS.peek(aDEVS.IPorts[I])))
        print "\t  New State: %s" % str(aDEVS.state)
        print "\t  Next scheduled internal transition at time %s" % `aDEVS.timeNext`

      ### ADDED 01/09/2006
      if __TRACE__:
        portInfo = ""
        for I in range(0, len(aDEVS.IPorts) ):
          portInfo += "<port name=\""+ aDEVS.IPorts[I].getPortName()+"\" category=\"I\">\n"
          portInfo += "<message>%s</message>\n</port>\n" % aDEVS.peek(aDEVS.IPorts[I])
        saveTrace(aDEVS.getModelFullName(), str(aDEVS.timeLast), "EX", portInfo, aDEVS.state.toXML(), str(aDEVS.state))

      ### ADDED 21/09/2011
      if __VCD__:
        for I in range(len(aDEVS.IPorts)):
            portName = aDEVS.IPorts[I].getPortName()
            portSignal = aDEVS.peek(aDEVS.IPorts[I])
            if portSignal != None:
                saveVCD(aDEVS.getModelFullName(), portName, str(portSignal))

#        saveTrace(aDEVS.getModelFullName(), str(aDEVS.timeLast), "external", aDEVS.state.toXML())
    # $(i,\,t)$ message --- sets origin of time at {\tt t}:
    elif msg[0] == 0:
      #aDEVS.send = AtomicSolver.receive
      #aDEVS.send = functools.partial(AtomicSolver.receive, self, aDEVS)
      aDEVS.timeLast = t - aDEVS.elapsed
      ta = aDEVS.timeAdvance()
      # Take into account the epsilon for floating point errors
      if ta < 0:
        raise DEVSException("Negative time advance in atomic model '" + \
                                aDEVS.getModelFullName() + "' with value " + \
                                str(ta) + " at time " + str(t))
      aDEVS.timeNext = aDEVS.timeLast + ta


      ### ADDED 01/09/2006
      if __TRACE__:
        saveTrace(aDEVS.getModelFullName(), str(aDEVS.timeLast), "EX", "", aDEVS.state.toXML(), str(aDEVS.state))

      ### ADDED 21/09/2011: code to save to VCD
      if __VCD__:
        for I in range(len(aDEVS.OPorts)):
            varList.append(VCDRecord(len(varList) + 1, aDEVS.getModelFullName(), aDEVS.OPorts[I].getPortName()))
        for I in range(len(aDEVS.IPorts)):
            varList.append(VCDRecord(len(varList) + 1, aDEVS.getModelFullName(), aDEVS.IPorts[I].getPortName()))

     
    else:
      raise DEVSException("Unrecognized message on atomic model '" + \
                          aDEVS.getModelFullName() + "'", 1)

#    ===================================================================    #

class CoupledSolver:
  """Simulator (coordinator) for coupled-DEVS.
  
  Coupled-DEVS can receive the same three types of messages as for atomic-
  DEVS, plus the $(y,\,t)$ message. The latter is implemented as a returned
  value rather than a message. This is possible because the $(y,\,t)$ message
  is always sent to a coupled-DEVS in response from its sending a $(*,\,t)$
  message. (This implementation makes it possible to easily distinguish
  $(y,\,t)$ from $(x,\,t)$ messages... to be completed)
  Description of {\tt eventList} and {\tt dStar} (to be completed)
  """

  def flatSelect(self, immChildrenOri):
        immChildren = list(immChildrenOri)
        # Check if there was a need for the EPSILON, as we will have to resort on the auto_id in this case
        if immChildren[0].timeNext != immChildren[-1].timeNext:
          immChildren.sort(key=lambda i : i.auto_id)

        # No need to sort on hierarchy, since the getModelFullName is originally used for the list
        # This name takes in account the order of the hierarchy
        # Not completely certain though...

        level = 0
        while (len(immChildren) > 1):
                if immChildren[0].hierarchy[level] == immChildren[-1].hierarchy[level]:
                        # All immChildren still agree on this level, iterate deeper
                        level += 1
                else:
                        # Disagreement in this level, so all immChildren need to get chosen
                        # on the previous level (that is at 'level - 1')
                        colLevel = level - 1
                        # Now we need to create the list that should be used for the select function
                        colChildren = []
                        last_id = -1
                        for i in immChildren:
                                new_id = i.hierarchy[level]
                                if new_id != last_id:
                                        # Select the coupled models that are colliding, not the real children
                                        # this is needed since the select function might try accessing these
                                        # coupled models to check for name etc. This way, the select function
                                        # should be able to be complex
                                        colChildren.append(id_fetch[new_id][3])
                                        last_id = new_id
                        cDEVS = id_fetch[colChildren[0].hierarchy[colLevel]][3]

                        # Only need the id for the moment
                        result_id = cDEVS.select(colChildren).auto_id
                        if __VERBOSE__:
                                printCollisionInfo(cDEVS, colChildren, id_fetch[result_id][3])
                        newChildren = []
                        for i in immChildren:
                                if i.hierarchy[level] == result_id:
                                        newChildren.append(i)
                        immChildren = newChildren
        return immChildren[0]
   
  def processExt(self, t, cDEVS, Y):
      """
        Define the processing of external events once and for all, this is a small slow-down due to an extra
        function call, but it is a lot more maintainable, certainly now that this function's size has exploded
      """
      elems = {}
      # All output ports

      for outport in Y:
           # For each of them, select the output component of it
           for inport in outport.outLine:
                # Now fetch the model that receives this message
                d = inport.hostDEVS.auto_id
                if d not in elems:
                        # Make the entry if this is the first data that the model receives
                        elems[d] = {}

                # Y.keys() is usually small enough to make a 'x in y' efficient enough
                # For speed, try to invoke a user-defined copy function
                try:
                        # User is willing to take full responsibility for copying
                        elems[d][inport] = Y[outport].copy()
                except AttributeError:
                        # Out of luck, do an expensive deepcopy
                        elems[d][inport] = deepcopy(Y[outport])

      # Now we only need to consider the models that actually receive input
      for i in elems:
           d = id_fetch[i][3]
           #print("Sending message to " + str(d.getModelFullName()))
           if (not __DIRECTCONNECT__) and (d.parent != cDEVS):
                # Not our responsibility to give that d it's data
                continue
           # Fetch this model's complete input
           X = elems[i]
           prevtime = d.timeNext
           # Send this model it's input
           #self.send(d, [X, t])
           d.send([X, t])
           ###

           # Check whether there was already an entry in the scheduler for this model
           # It might be that the element was passive until it received input!
           if prevtime != float('inf'):
                   # Just invalidate this event
                   cDEVS.invalids += 1

           # First the event is fetched and made invalid
           event = id_fetch[d.auto_id]
           event[2] = False

           # Now a completely new 'unrelated' event is created, to remove any connection
           # to the previously scheduled (now invalidated) event
           # By overwriting the entry in id_fetch, we now always use the new entry
           id_fetch[d.auto_id] = [d.timeNext, event[1], True, event[3]]

           # Only when the event doesn't passivate, the heap should be updated
           # All the previous work was still needed though, since we had to destroy
           # the references to the unscheduled event
           if d.timeNext != float('inf'):
                   heappush(cDEVS.heap, id_fetch[d.auto_id])

      # Check whether cleanup is required: this is the expensive part of the heap!
      # NOTE:   check whether there are too many invalids in the heap. If this occurs, we should
      #         clean everything. This is a rather expensive operation (O(n)), but is required to
      #         keep an efficient heap usage by reducing the heap size.
      #         This threshold might be adjusted to increase performance, depending on the situation. 
      #         As long as it gets called before we run out of memory.
      if cDEVS.invalids >= 2 * len(cDEVS.componentSet):
        newheap = []
        for i in cDEVS.heap:
                # If still valid and not infinite: keep it
                if i[2] and (i[0] != float('inf')):
                        newheap.append(i)
        cDEVS.heap = newheap

        # Now we still need to make a valid heap of it, since the invalids were an important part of the heap
        heapify(cDEVS.heap)

        # Obviously, there are no more invalids
        cDEVS.invalids = 0

      try:
        while not cDEVS.heap[0][2]:
                heappop(cDEVS.heap)
                cDEVS.invalids -= 1
        cDEVS.timeNext = cDEVS.heap[0][0]
      except IndexError:
        cDEVS.timeNext = float('inf')

  def receive(self, cDEVS, msg):
    # For any received message, the time {\tt t} (time at which the message 
    # is sent) is the second item in the list {\tt msg}.
    t = msg[1]
    
    # $(*,\,t)$ message --- triggers internal transition and returns
    # $(y,\,t)$ message for parent coupled-DEVS:
    if msg[0] == 1:

      if (t - cDEVS.timeNext) >= EPSILON:
        raise DEVSException("Bad synchronization in coupled model '" + \
                            cDEVS.getModelFullName() + "' on internal " + \
                            "transition; simulation time = " + str(t) + \
                            "; model.timeNext = " + str(cDEVS.timeNext), 1)

      immChildren = []

      if __LAZY__:
        for i in cDEVS.prevCollide:
                immChildren.append(i[3])
        # Now clear the list
        cDEVS.prevCollide = []

      # Find all imminent children still in the heap, this might very well result in 0 found if lazyness is used!
      try:
        while (abs(cDEVS.heap[0][0] - t) < EPSILON):
                # Check if the found event is actually still active
                if(cDEVS.heap[0][2]) and ((not __LAZY__) or (cDEVS.heap[0][3] not in immChildren)):
                        # Active, so event is imminent
                        immChildren.append(cDEVS.heap[0][3])
                else:
                        # Wasn't active, but we will have to pop this to get the next
                        # So we can lower the number of invalids
                        cDEVS.invalids -= 1
                # Advance the while loop
                heappop(cDEVS.heap)
      except IndexError:
        # It seems the end of the heap is reached, so all scheduled events are done
        # Nothing do do anymore
        pass
      #print("Got immchildren: " + str(immChildren))

      # NOTE: compared to the original implementation, no sort of these imminent children
      #       is needed since the heap guarantees a sorted order of pops
      #       This is the reason for the order of the data in events
      #       [ timeNext, auto_id, valid, event]
      #         --> first sort on timeNext (logical)
      #         --> then on auto_id, which is based on the full model name: guarantees determinism

      if __VCD__:
        # Convert float to integer without losing precision
        # ex. 5.0 --> 50, 5.5 --> 55
        global PREVIOUSTIMESTAMP
        vcdTime = int(str(int(floor(t))) + str(int(t - floor(t)) * (len(str(t)) - 2)))

        if (PREVIOUSTIMESTAMP != vcdTime):
           # The time has passed, so add a new VCD header
           vcdList.append("#" + str(vcdTime))
           PREVIOUSTIMESTAMP = vcdTime

      # Now fetch the model that should be processed
      if len(immChildren) == 1:
        # Easy, only one possibility
        dStar = immChildren[0]
      ###  ADDED 04/04/05: helpful to debug
      elif len(immChildren) == 0:
        dStar = None
      ###
      else:  # len(immChildren) > 1
        # Invoke the select function of the coupled model

        # This is the statement to execute without direct connection
        #dStar = cDEVS.select(immChildren)

        # This is the one to execute with direct connection
        if __DIRECTCONNECT__: 
                dStar = self.flatSelect(immChildren)
        else:
                dStar = cDEVS.select(immChildren)
                # Print some info about the collision
                if __VERBOSE__:
                        printCollisionInfo(cDEVS, immChildren, dStar)

        if not __LAZY__:
                for i in immChildren:
                        # Don't reinsert the chosen model
                        if i != dStar:
                                heappush(cDEVS.heap, id_fetch[i.auto_id])

      # Send $(*,\,t)$ message to active children, which returns (or sends
      # back) message $(y,\,t)$. In the present implementation, just the
      # sub-DEVS output dictionary $y$ is returned and stored in {\tt Y}:
      #Y = self.send(dStar, msg)
      Y = dStar.send(msg)
      ###
      
      # Scan the list of Output Ports to find those influenced by (connected to)
      # the active child. Don't send these messages all at once, since it might be
      # possible that an element receives input on two different ports. The DEVS
      # formalisms allows us to send two different events, though this would be
      # highly inefficient.
      cDEVS.timeLast = t

      # If the model passivates, no rescheduling is required thus reducing the size of the heap
      if dStar.timeNext != float('inf'):
        # Model doesn't passivates
        id_fetch[dStar.auto_id][0] = dStar.timeNext
        heappush(cDEVS.heap, id_fetch[dStar.auto_id])

      self.processExt(t, cDEVS, Y)

      if len(immChildren) > 1:
        # Now reinsert all these models that were previously popped
        # NOTE: this is a disadvantage of Classic DEVS compared to Parallel DEVS
        #       since PDEVS just executes all of them, while CDEVS requires only
        #       one model to be executed. It might be possible to save these immChildren
        #       though this might cause a lot of overhead, since the dStar might
        #       influence (and thus reschedule) the immChildren. This would cause
        #       not only updates to the heap, but also to the immChildren
        # NOTE: try to solve it with keeping a temporary list, which is checked first,
        #       thus removing the need to repush/repop the same elements
        if __LAZY__:
                mintime = float('inf')
                for i in immChildren:
                        if i != dStar:
                                record = id_fetch[i.auto_id]
                                if abs(record[0] - t) < EPSILON and record[2]:
                                        cDEVS.prevCollide.append(record)
                                        mintime = min(mintime, record[0])
                if len(cDEVS.prevCollide) > 0:
                        # Reset the timeNext to the currentTime if there is a valid one in the list
                        cDEVS.timeNext = mintime
       

      # Use same pattern as above to update coupled-DEVS output dictionary
      # (once again using ports' {\tt inLine} attributes rather than
      # coupled-DEVS' {\tt EOC}). The DEVS' output is then returned to its
      # parent coupled-DEVS (rather than being sent).
      myOutput = {}
      for p in cDEVS.OPorts:
        for pi in p.inLine:
         try:
                try:
                        myOutput[p] = Y[pi].copy()
                except AttributeError:
                        myOutput[p] = deepcopy(Y[pi])
                break
         except KeyError:
                pass
      cDEVS.myOutput = myOutput

      return cDEVS.myOutput

    
    # ${x,\,t)$ message --- triggers external transition, where $x$ is the
    # input dictionary to the DEVS:
    elif type(msg[0]) == type({}):
      if not(cDEVS.timeLast - EPSILON <= t <= cDEVS.timeNext + EPSILON):
        raise DEVSException("Bad synchronization in coupled model '" + \
                            cDEVS.getModelFullName() + "' on external " + \
                            "transition; simulation time = " + str(t) + \
                            "; model.timeLast = " + str(cDEVS.timeLast) + \
                            "; model.timeNext = " + str(cDEVS.timeNext), 1)
      
      cDEVS.myInput = msg[0]      

      # Send $(x,\,t)$ messages to those sub-DEVS influenced by the coupled-
      # DEVS input ports (ref. {\tt EIC}). This is essentially the same code
      # as above that also updates the coupled-DEVS' time variables in
      # parallel.
      cDEVS.timeLast = t

      self.processExt(t, cDEVS, cDEVS.myInput)
    # $(i,\,t)$ message --- sets origin of time at {\tt t}:
    elif msg[0] == 0:
      
      # Rebuild event-list and update time variables, by sending the
      # initialization message to all the children of the coupled-DEVS. Note
      # that the event-list is not sorted here, but only when the list of
      # {\sl imminent children\/} is needed. Also note that {\tt None} is
      # defined as bigger than any number in Python (stands for $+\infty$).
      # >> HV: now that we have INFINITY, should that not be used ?
      # Use the python float('inf') instead
      cDEVS.timeLast = 0.
      cDEVS.timeNext = float('inf')
      cDEVS.heap = []
      cDEVS.prevCollide = []
      cDEVS.invalids = 0
      # This part isn't fast, but it doesn't matter, since it just inits everything, optimizing here doesn't
      # matter as it is only called once AND every element has to be initted.
      for d in cDEVS.componentSet:
        #self.send(d, msg)
        if d.type() == 'COUPLED':
                d.send = functools.partial(CoupledSolver.receive, self, d)
        elif d.type() == 'ATOMIC':
                d.send = functools.partial(AtomicSolver.receive, self, d)
        d.send(msg)
        ###
        cDEVS.timeNext = min(cDEVS.timeNext, d.timeNext)
        cDEVS.timeLast = max(cDEVS.timeLast, d.timeLast)
        id_fetch[d.auto_id][0] = d.timeNext
        if d.timeNext != float('inf'):
                # Setup an original heap schedule
                heappush(cDEVS.heap, id_fetch[d.auto_id])
       
 
    else:
      raise DEVSException("Unrecognized message on atomic model '" + \
                          cDEVS.getModelFullName() + "'", 1)

#    ===================================================================    #

class Simulator(AtomicSolver, CoupledSolver):
  """Associates a hierarchical DEVS model with the simulation engine.
  """
  
  ###
  def __init__(self, model):
    """Constructor.
  
    {\tt model} is an instance of a valid hierarchical DEVS model. The
    constructor stores a local reference to this model and augments it with
    {\sl time variables\/} required by the simulator.
    """
  
    self.model = model
    
    self.augment(self.model)
  
  ###
  def augment(self, d):
    """Recusively augment model {\tt d} with {\sl time variables\/}.
    """
   
    # {\tt timeLast} holds the simulation time when the last event occured
    # within the given DEVS, and (\tt timeNext} holds the scheduled time for
    # the next event.
    d.timeLast = d.timeNext = 0.
    if d.type() == 'COUPLED':
      # {\tt eventList} is the list of pairs $(tn_d,\,d)$, where $d$ is a
      # reference to a sub-model of the coupled-DEVS and $tn_d$ is $d$'s
      # time of next event.
      for subd in d.componentSet:
        self.augment(subd)
        
  ###
  def constructNameList(self, model, nameList):
    nameList.append([model.getModelFullName(), model])
    if model.type() == "ATOMIC":
      return
    for i in model.componentSet:
      self.constructNameList(i, nameList)

  ###
  def directConnect(self):
      """
      Perform the direct connecting, this step is optional though suggested for very deeply nested models, all coupled
      models will be expanded to the root layer, potentially speeding up the simululation due to reduced message passing.
      This algorithm might take some time for complex couplings, though it will speed it up in the long run
      """
      ### Generate the hierarchy lists
      self.model.setHierarchy([])
      ### Now actually reconnect the model
      newlist = []
      for i in self.model.componentSet:
         if i.type() == "COUPLED":
                self.model.componentSet.extend(i.componentSet)
         else:
                newlist.append(i)
      self.model.componentSet = newlist

      ### Fix all ports: cumbersome process
      for i in self.model.componentSet:
         # Remap the output ports
         for outport in i.OPorts:
                # The new contents of the line
                newLine = []
                for outline in outport.outLine:
                        # If it is a coupled model, we must expand this model
                        if outline.hostDEVS.type() == 'COUPLED':
                                for inline in outline.outLine:
                                    # Add it to the current iterating list, so we can just continue
                                    outport.outLine.append(inline)
                                    # If it is a Coupled model, we should just continue expanding it and not add it to the finished line
                                    if inline.hostDEVS.type() != "COUPLED":
                                         newLine.append(inline)
                        elif outline not in newLine:
                                # Add to the new line if it isn't already there
                                # Note that it isn't really mandatory to check for this, it is a lot cleaner to do so
                                # This will greatly increase the complexity of the connector though
                                newLine.append(outline)
                outport.outLine = newLine
         # Remap the input ports: identical to the output ports, only in the reverse direction
         for inport in i.IPorts:
                newLine = []
                for inline in inport.inLine:
                        if inline.hostDEVS.type() == 'COUPLED':
                                for outline in inline.inLine:
                                    inport.inLine.append(outline)
                                    if outline.hostDEVS.type() != "COUPLED":
                                         newLine.append(outline)
                        elif inline not in newLine:
                                newLine.append(inline)
                inport.inLine = newLine

  ###
  def simulate(self,
               termination_condition=lambda model, clock: False,
               verbose=False, trace=False, vcd=False, dconnect=True, lazy=False, seed=1):
    """Simulate the model (Root-Coordinator).
    
    In this implementation, the simulation stops only when the simulation
    time (stored in {\tt clock}) reaches {\tt T} (starts at 0). Other means
    to stop the simulation ({\it e.g.}, when an atomic-DEVS enters a given
    state, or when a global variable reaches a predefined value) will be
    implemented.
    """

    global __VERBOSE__
    __VERBOSE__ = verbose

    # Added at 01/09/2006 by Bill Song
    global __TRACE__
    __TRACE__ = trace

    if __TRACE__:
      global traceList
      traceList = []

    # Added 21/09/2011 by Yentl Van Tendeloo
    global __VCD__
    __VCD__ = vcd
    if __VCD__:
      global vcdList
      vcdList = []

      # VCD needs a list of all variables/wires at the start
      global varList
      varList = []

    # Added 01/10/2012 by Yentl Van Tendeloo
    global __DIRECTCONNECT__
    __DIRECTCONNECT__ = dconnect
    if self.model.type() == "ATOMIC":
        __DIRECTCONNECT__ = False

    # Added 08/11/2012 by Yentl Van Tendeloo
    global __LAZY__
    __LAZY__ = lazy


    # Seed the random number generator
    # This needs to be done deterministically
    # to guarantee the same stream of pseudo-random
    # numbers for each (repeated) simulation experiment.
    # By default, the random number generated is
    # seeded by a number obtained from the system clock.
    #
    random.seed(seed)

    # Reset for the simulation
    global id_fetch
    id_fetch = []
    try:
      ##### Optimizations
      ### Use automatically generated ID's instead of getModelFullName
      self.nameList = []
      self.constructNameList(self.model, self.nameList)
      # The only real sort that is done, only happens once
      self.nameList.sort(key=lambda i:[i[0], i[1].myID])
      count = 0
      # Give them numbers
      for i in self.nameList:
        i[1].auto_id = count
        count += 1
        # NOTE: this way, id_fetch is a list and not a dict. A dict has a
        #       high amortized worst case lookup, and since the accesses
        #       are done based on an integer and a constant list without holes,
        #       nothing beats a good old list in this case
        id_fetch.append([i[1].timeNext, i[1].auto_id, True, i[1]])
      ##### Direct connection
      if __DIRECTCONNECT__:
        self.directConnect()

      # Initialize the model --- set the simulation clock to 0.
      if self.model.type() == "ATOMIC":
        self.model.send = functools.partial(AtomicSolver.receive, self, self.model)
      else:
        self.model.send = functools.partial(CoupledSolver.receive, self, self.model)
      #self.send(self.model, (0, 0))
      self.model.send([0, 0])
      ###
      clock = self.model.timeNext


      ### ADDED 06/06/05: code modified to display initial time:
      if __VERBOSE__:
        print "\n", "__  Current Time: %10.2f" % 0.0, "_" * 42, "\n"
        for i in id_fetch:
          model = i[3]
          printInitialCondition(model)

    # Main loop repeatedly sends $(*,\,t)$ messages to the model's root
    # DEVS.
      while not termination_condition(self.model, clock):
        if __VERBOSE__:
          print "\n", "__  Current Time: %10.2f" % clock, "_" * 42, "\n"
        #self.send(self.model, (1, clock))
        self.model.send([1, clock])
        ###

        if __VERBOSE__:
          if len(self.model.OPorts) > 0:
            # don't print if root model is autonomous
            # (no output ports)
            print "\n\tROOT DEVS <%s> Output Port Configuration:" \
                  % self.model.getModelFullName()
          for I in range(0, len(self.model.OPorts) ):
            if self.model.OPorts[I] in self.model.myOutput:
  #            print "BAR"
              print "\t    port <%s>: %s" % (\
                           self.model.OPorts[I].getPortName(), \
                           self.model.myOutput[self.model.OPorts[I]])
  #              self.model.myOutput[self.model.OPorts[I]], \
  #              repr(self.model.myOutput[self.model.OPorts[I]]))
            else:
              print "\t    port <%s>: NoEvent" % self.model.OPorts[I].getPortName()
        
        clock = self.model.timeNext
    except DEVSException, e:
        print e
        # Should also exit on a DEVSException since this isn't really meant to happen
        import sys
        sys.exit(1)

    if(__TRACE__):
      traceFile = open("devstrace.xml", "w")
      traceFile.write("<?xml version=\"1.0\"?>\n" + "<trace>\n" )
      for evt in traceList:
                                traceFile.write(evt)
      traceFile.write("</trace>")
      traceFile.close()

    if __VCD__:
      vcdFile = open("devstrace.vcd", "w")
      vcdFile.write(vcdHeader())
      vcdFile.write(vcdVariables())
      for evt in vcdList:
        vcdFile.write(evt)
        vcdFile.write("\n")
      # Prevent a double timestamp
      if "#" not in vcdList[-1]:
        vcdFile.write("#" + str(PREVIOUSTIMESTAMP + 10))
      vcdFile.close()

# Set to False to disable the use of CyDEVS...
__USE_CYDEVS__ = False
if __USE_CYDEVS__:
 try:
  from cydevs import *
 except ImportError:
  print("Couldn't load CyDEVS, falling back to PyDEVS")
  print("To disable this message, run the build.sh script, or set __USE_CYDEVS__ to False in simulator.py")
